/*
** Copyright (C) 1999 Erik de Castro Lopo <erikd@zip.com.au>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include	<stdio.h>
#include	<unistd.h>
#include	<string.h>
#include	<math.h>
#include	<ctype.h>

#include	"sndfile.h"
#include	"config.h"
#include	"sfendian.h"
#include	"common.h"
#include	"pcm.h"


enum /* Here is the list of known format tags */
{	
	WAVE_FORMAT_UNKNOWN				= 0x0000,		/* Microsoft Corporation */
	WAVE_FORMAT_PCM     		    = 0x0001, 		/* Microsoft PCM format */

	WAVE_FORMAT_ADPCM				= 0x0002,		/* Microsoft ADPCM */
	WAVE_FORMAT_IEEE_FLOAT			= 0x0003,		/* Micrososft 32 bit float format */
	
	WAVE_FORMAT_IBM_CVSD			= 0x0005,		/* IBM Corporation */
	WAVE_FORMAT_ALAW				= 0x0006,		/* Microsoft Corporation */
	WAVE_FORMAT_MULAW				= 0x0007,		/* Microsoft Corporation */
	WAVE_FORMAT_OKI_ADPCM			= 0x0010,		/* OKI */
	WAVE_FORMAT_IMA_ADPCM			= 0x0011,		/* Intel Corporation */
	WAVE_FORMAT_MEDIASPACE_ADPCM	= 0x0012,		/* Videologic */
	WAVE_FORMAT_SIERRA_ADPCM		= 0x0013,		/* Sierra Semiconductor Corp */
	WAVE_FORMAT_G723_ADPCM			= 0x0014,		/* Antex Electronics Corporation */
	WAVE_FORMAT_DIGISTD				= 0x0015,		/* DSP Solutions, Inc. */
	WAVE_FORMAT_DIGIFIX				= 0x0016,		/* DSP Solutions, Inc. */
	WAVE_FORMAT_DIALOGIC_OKI_ADPCM	= 0x0017,		/*  Dialogic Corporation  */
	WAVE_FORMAT_MEDIAVISION_ADPCM	= 0x0018,		/*  Media Vision, Inc. */

	WAVE_FORMAT_YAMAHA_ADPCM		= 0x0020,		/* Yamaha Corporation of America */
	WAVE_FORMAT_SONARC				= 0x0021,		/* Speech Compression */
	WAVE_FORMAT_DSPGROUP_TRUESPEECH = 0x0022,		/* DSP Group, Inc */
	WAVE_FORMAT_ECHOSC1				= 0x0023,		/* Echo Speech Corporation */
	WAVE_FORMAT_AUDIOFILE_AF18  	= 0x0024,		/* Audiofile, Inc. */
	WAVE_FORMAT_APTX				= 0x0025,		/* Audio Processing Technology */
	WAVE_FORMAT_AUDIOFILE_AF10  	= 0x0026,		/* Audiofile, Inc. */

	WAVE_FORMAT_DOLBY_AC2			= 0x0030,		/* Dolby Laboratories */
	WAVE_FORMAT_GSM610				= 0x0031,		/* Microsoft Corporation */
	WAVE_FORMAT_MSNAUDIO			= 0x0032,		/* Microsoft Corporation */
	WAVE_FORMAT_ANTEX_ADPCME		= 0x0033, 		/* Antex Electronics Corporation */
	WAVE_FORMAT_CONTROL_RES_VQLPC	= 0x0034,		/* Control Resources Limited */
	WAVE_FORMAT_DIGIREAL			= 0x0035,		/* DSP Solutions, Inc. */
	WAVE_FORMAT_DIGIADPCM			= 0x0036,		/* DSP Solutions, Inc. */
	WAVE_FORMAT_CONTROL_RES_CR10	= 0x0037,		/* Control Resources Limited */
	WAVE_FORMAT_NMS_VBXADPCM		= 0x0038,		/* Natural MicroSystems */
	WAVE_FORMAT_ROCKWELL_ADPCM		= 0x003B,		/* Rockwell International */
	WAVE_FORMAT_ROCKWELL_DIGITALK	= 0x003C, 		/* Rockwell International */

	WAVE_FORMAT_G721_ADPCM			= 0x0040,		/* Antex Electronics Corporation */
	WAVE_FORMAT_MPEG				= 0x0050,		/* Microsoft Corporation */

	WAVE_FORMAT_MPEGLAYER3			= 0x0055,		/* MPEG 3 Layer 1 */

	IBM_FORMAT_MULAW				= 0x0101,		/* IBM mu-law format */
	IBM_FORMAT_ALAW					= 0x0102,		/* IBM a-law format */
	IBM_FORMAT_ADPCM				= 0x0103,		/* IBM AVC Adaptive Differential PCM format */

	WAVE_FORMAT_CREATIVE_ADPCM		= 0x0200,		/* Creative Labs, Inc */

	WAVE_FORMAT_FM_TOWNS_SND		= 0x0300,		/* Fujitsu Corp. */
	WAVE_FORMAT_OLIGSM				= 0x1000,		/* Ing C. Olivetti & C., S.p.A. */
	WAVE_FORMAT_OLIADPCM			= 0x1001,		/* Ing C. Olivetti & C., S.p.A. */
	WAVE_FORMAT_OLICELP				= 0x1002,		/* Ing C. Olivetti & C., S.p.A. */
	WAVE_FORMAT_OLISBC				= 0x1003,		/* Ing C. Olivetti & C., S.p.A. */
	WAVE_FORMAT_OLIOPR				= 0x1004,		/* Ing C. Olivetti & C., S.p.A. */

	WAVE_FORMAT_EXTENSIBLE			= 0xFFFE
} ;

typedef	struct
{	unsigned short	format ;
	unsigned short	channels ;
	unsigned int	samplerate ;
	unsigned int	bytespersec ;
	unsigned short	blockalign ;
	unsigned short	bitwidth ;
} MIN_WAV_FMT ;

typedef	struct
{	unsigned short	format ;
	unsigned short	channels ;
	unsigned int	samplerate ;
	unsigned int	bytespersec ;
	unsigned short	blockalign ;
	unsigned short	bitwidth ;
	unsigned short	extrabytes ;
	unsigned short	samplesperblock ;
} IMA_WAV_FMT ;

typedef union
{	unsigned short	format ;
	MIN_WAV_FMT		min ;
	IMA_WAV_FMT		ima ;
	char			padding [256] ;
} WAV_FMT ;

typedef	struct
{	int		index, prev ;
} IMA_STATE ;

typedef struct
{	int				channels, blocksize, samplesperblock, blocks ; 
	int				blockcount, samplecount ;
	int				previous [2] ;
	int				stepindex [2] ;
	unsigned char	*block ;
	short			*samples ;
	unsigned char	data	[4] ; /* Dummy size */
} IMA_PRIVATE ;

/*------------------------------------------------------------------------------
 * Macros to handle big/little endian issues.
 */

#if (__LITTLE_ENDIAN__ == 1)
#	define	MAKE_MARKER(a,b,c,d)		((a)|((b)<<8)|((c)<<16)|((d)<<24))
#elif (__BIG_ENDIAN__ == 1)
#	define	MAKE_MARKER(a,b,c,d)		(((a)<<24)|((b)<<16)|((c)<<8)|(d))
#else
#	error "Cannot determine endian-ness of processor."
#endif

#define RIFF_MARKER	(MAKE_MARKER ('R', 'I', 'F', 'F')) 
#define WAVE_MARKER	(MAKE_MARKER ('W', 'A', 'V', 'E')) 
#define fmt_MARKER	(MAKE_MARKER ('f', 'm', 't', ' ')) 
#define data_MARKER	(MAKE_MARKER ('d', 'a', 't', 'a')) 
#define cue_MARKER	(MAKE_MARKER ('c', 'u', 'e', ' ')) 
#define LIST_MARKER	(MAKE_MARKER ('L', 'I', 'S', 'T')) 
#define slnt_MARKER	(MAKE_MARKER ('s', 'l', 'n', 't')) 
#define wavl_MARKER	(MAKE_MARKER ('w', 'a', 'v', 'l')) 
#define INFO_MARKER	(MAKE_MARKER ('I', 'N', 'F', 'O')) 
#define plst_MARKER	(MAKE_MARKER ('p', 'l', 's', 't')) 
#define adtl_MARKER	(MAKE_MARKER ('a', 'd', 't', 'l')) 
#define labl_MARKER	(MAKE_MARKER ('l', 'a', 'b', 'l')) 
#define note_MARKER	(MAKE_MARKER ('n', 'o', 't', 'e')) 
#define fact_MARKER	(MAKE_MARKER ('f', 'a', 'c', 't')) 
#define smpl_MARKER	(MAKE_MARKER ('s', 'm', 'p', 'l')) 
#define bext_MARKER	(MAKE_MARKER ('b', 'e', 'x', 't')) 
#define MEXT_MARKER	(MAKE_MARKER ('M', 'E', 'X', 'T')) 
#define DISP_MARKER	(MAKE_MARKER ('D', 'I', 'S', 'P')) 

/*------------------------------------------------------------------------------
 * Private static functions.
 */

static int		__wav_close	(SF_PRIVATE  *psf) ;
static int		ima_close	(SF_PRIVATE  *psf) ;

static int		read_fmt_chunk	(SF_PRIVATE *psf, WAV_FMT *wav_fmt) ;
static int		write_header_fmt (SF_PRIVATE *psf, WAV_FMT *wav_fmt, int size) ;

static const 	char* wav_format_str (int k) ;

static int		wav_read_x86f2s (SF_PRIVATE *psf, short *ptr, int len) ;
static int		wav_read_x86f2i (SF_PRIVATE *psf, int *ptr, int len) ;
static int		wav_read_x86f2d (SF_PRIVATE *psf, double *ptr, int len, int normalize) ;

static int		wav_write_s2x86f (SF_PRIVATE *psf, short *ptr, int len) ;
static int		wav_write_i2x86f (SF_PRIVATE *psf, int *ptr, int len) ;
static int		wav_write_d2x86f (SF_PRIVATE *psf, double *ptr, int len, int normalize) ;

static int		ima_reader_init (SF_PRIVATE *psf, WAV_FMT *fmt) ;
static int		ima_read_s (SF_PRIVATE *psf, short *ptr, int len) ;
static int		ima_read_i (SF_PRIVATE *psf, int *ptr, int len) ;
static int		ima_read_d (SF_PRIVATE *psf, double *ptr, int len, int normalize) ;
static off_t    ima_seek   (SF_PRIVATE *psf, off_t offset, int whence) ;

static int		ima_writer_init (SF_PRIVATE *psf, WAV_FMT *fmt) ;
static int		ima_write_s (SF_PRIVATE *psf, short *ptr, int len) ;
static int		ima_write_i (SF_PRIVATE *psf, int *ptr, int len) ;
static int		ima_write_d (SF_PRIVATE *psf, double *ptr, int len, int normalize) ;

static 
void	le2h_wav_fmt (WAV_FMT *fmt)
{	
	fmt->min.format      = ENDSWAP_SHORT (fmt->min.format) ;
	fmt->min.channels    = ENDSWAP_SHORT (fmt->min.channels) ;
	fmt->min.samplerate  = ENDSWAP_INT   (fmt->min.samplerate) ;
	fmt->min.bytespersec = ENDSWAP_INT   (fmt->min.bytespersec) ;
	fmt->min.blockalign  = ENDSWAP_SHORT (fmt->min.blockalign) ;
	fmt->min.bitwidth    = ENDSWAP_SHORT (fmt->min.bitwidth) ;

	switch (fmt->format)
	{	case WAVE_FORMAT_IMA_ADPCM :	
				fmt->ima.extrabytes      = ENDSWAP_SHORT (fmt->ima.extrabytes) ;
				fmt->ima.samplesperblock = ENDSWAP_SHORT (fmt->ima.samplesperblock) ;
				break ;
				
		default : break ;
		} ;

} /* le2h_wav_fmt */

static 
void	h2le_wav_fmt (WAV_FMT *fmt)
{	
	switch (fmt->format)
	{	case WAVE_FORMAT_IMA_ADPCM :	
				fmt->ima.extrabytes      = ENDSWAP_SHORT (fmt->ima.extrabytes) ;
				fmt->ima.samplesperblock = ENDSWAP_SHORT (fmt->ima.samplesperblock) ;
				break ;
				
		default : break ;
		} ;

	fmt->min.format      = ENDSWAP_SHORT (fmt->min.format) ;
	fmt->min.channels    = ENDSWAP_SHORT (fmt->min.channels) ;
	fmt->min.samplerate  = ENDSWAP_INT   (fmt->min.samplerate) ;
	fmt->min.bytespersec = ENDSWAP_INT   (fmt->min.bytespersec) ;
	fmt->min.blockalign  = ENDSWAP_SHORT (fmt->min.blockalign) ;
	fmt->min.bitwidth    = ENDSWAP_SHORT (fmt->min.bitwidth) ;

} /* h2le_wav_fmt */

/*------------------------------------------------------------------------------
 * Public functions.
 */

int 	__wav_open_read	(SF_PRIVATE *psf)
{	WAV_FMT			wav_fmt ;
	unsigned int	dword, marker, RIFFsize ;
	int				parsestage = 0, error, format = 0 ;
	
	while (1)
	{	read (psf->fd, &marker, sizeof (marker)) ;
		switch (marker)
		{	case RIFF_MARKER :
					if (parsestage != 0)
						return SFE_WAV_NO_RIFF ;
					read (psf->fd, &dword, sizeof (dword)) ;
					RIFFsize = LE2H_INT (dword) ;	
					if (psf->filelength  < RIFFsize + 2 * sizeof (dword))
					{	dword = psf->filelength - 2 * sizeof (dword);
						__psf_sprintf (psf, "RIFF : %d (should be %d)\n", RIFFsize, dword) ;
						RIFFsize = dword ;
						}
					else
						__psf_sprintf (psf, "RIFF : %d\n", RIFFsize) ;
					parsestage = 1 ;
					break ;
					
			case WAVE_MARKER :
					if (parsestage != 1)
						return SFE_WAV_NO_WAVE ;
					__psf_sprintf (psf, "WAVE\n") ;
					parsestage = 2 ;
					break ;
			
			case fmt_MARKER :
					if (parsestage != 2)
						return SFE_WAV_NO_FMT ;
					if ((error = read_fmt_chunk (psf, &wav_fmt)))
						return error ;
						
					format     = wav_fmt.format ;
					parsestage = 3 ;
					break ;
					
			case data_MARKER :
					if (parsestage < 3)
						return SFE_WAV_NO_DATA ;
					read (psf->fd, &dword, sizeof (dword)) ;
					psf->datalength = LE2H_INT (dword) ;
					psf->dataoffset = lseek (psf->fd, 0, SEEK_CUR) ;
					
					if (psf->filelength < psf->dataoffset + psf->datalength)
						__psf_sprintf (psf, "data : %d (should be %d)\n", psf->datalength, psf->filelength - psf->dataoffset) ;
					else
						__psf_sprintf (psf, "data : %d\n", psf->datalength) ;

					lseek (psf->fd, psf->datalength, SEEK_CUR) ;
					break ;

			case cue_MARKER :
			case LIST_MARKER :
			case INFO_MARKER :
			case fact_MARKER :
			case smpl_MARKER :
			case bext_MARKER :
			case MEXT_MARKER :
			case DISP_MARKER :
					read (psf->fd, &dword, sizeof (dword)) ;
					__psf_sprintf (psf, "%D : %d\n", marker, dword) ;
					lseek (psf->fd, dword, SEEK_CUR) ;
					break ;

			default : 
#if 0
					if (isprint ((marker >> 24) & 0xFF) && isprint ((marker >> 16) & 0xFF)
						&& isprint ((marker >> 8) & 0xFF) && isprint (marker & 0xFF))
#endif
					{
						read (psf->fd, &dword, sizeof (dword)) ;
						__psf_sprintf (psf, "%D : %d (unknown marker)\n", marker, dword) ;
						lseek (psf->fd, dword, SEEK_CUR) ;
						break ;
						} ;
					__psf_sprintf (psf, "*** Unknown chunk marker : %D\n", marker) ;
					return SFE_WAV_UNKNOWN_CHUNK ;
			} ;	/* switch (dword) */

		if (lseek (psf->fd, 0, SEEK_CUR) >= (int) (psf->filelength - 8))
			break ;
		} ; /* while (1) */
		
	if (! psf->dataoffset)
		return SFE_WAV_NO_DATA ;

	psf->current     = 0 ;
	psf->endian      = SF_ENDIAN_LITTLE ;		/* All WAV files are little endian. */
	psf->sf.seekable = SF_TRUE ;
	psf->sf.sections = 1 ;
	
	lseek (psf->fd, psf->dataoffset, SEEK_SET) ;
	
	psf->close = (func_close) __wav_close ;

	if (format == WAVE_FORMAT_IMA_ADPCM)
	{	psf->sf.format   = (SF_FORMAT_WAV | SF_FORMAT_IMA_ADPCM) ;
		if ((error = ima_reader_init (psf, &wav_fmt)))
			return error ;
		psf->read_short  = (func_rs) ima_read_s ;
		psf->read_int    = (func_ri) ima_read_i ;
		psf->read_double = (func_rd) ima_read_d ;
		psf->seek_func   = (func_seek) ima_seek ;
		return 0 ;
		} ;

	if (psf->blockwidth)
	{	if (psf->filelength - psf->dataoffset < psf->datalength)
			psf->sf.samples = (psf->filelength - psf->dataoffset) / psf->blockwidth ;
		else
			psf->sf.samples = psf->datalength / psf->blockwidth ;
		} ;
		
	if (format == WAVE_FORMAT_IEEE_FLOAT)
	{	if (CAN_READ_WRITE_x86_IEEE)
		{	psf->sf.format   = (SF_FORMAT_WAV | SF_FORMAT_FLOAT) ;
			psf->read_short  = (func_rs) __pcm_read_f2s ;
			psf->read_int    = (func_ri) __pcm_read_f2i ;
			psf->read_double = (func_rd) __pcm_read_f2d ;
			}
		else 
		{	psf->sf.format   = (SF_FORMAT_WAV | SF_FORMAT_FLOAT) ;
			psf->read_short  = (func_rs) wav_read_x86f2s ;
			psf->read_int    = (func_ri) wav_read_x86f2i ;
			psf->read_double = (func_rd) wav_read_x86f2d ;
			} ;
		return	 0 ;
		} ;
	
	if (format != WAVE_FORMAT_PCM)
		return SFE_UNIMPLEMENTED ;

	psf->sf.format = (SF_FORMAT_WAV | SF_FORMAT_PCM) ;
	switch (psf->bytewidth)
	{	case  1 :
				psf->read_short  = (func_rs) __pcm_read_uc2s ;
				psf->read_int    = (func_ri) __pcm_read_uc2i ;
				psf->read_double = (func_rd) __pcm_read_uc2d ;
				break ;
		case  2 :
				psf->read_short  = (func_rs) __pcm_read_les2s ;
				psf->read_int    = (func_ri) __pcm_read_les2i ;
				psf->read_double = (func_rd) __pcm_read_les2d ;
				break ;
		case  3 :
				psf->read_short  = (func_rs) __pcm_read_let2s ;
				psf->read_int    = (func_ri) __pcm_read_let2i ;
				psf->read_double = (func_rd) __pcm_read_let2d ;
				break ;
		case  4 :
				psf->read_short  = (func_rs) __pcm_read_lei2s ;
				psf->read_int    = (func_ri) __pcm_read_lei2i ;
				psf->read_double = (func_rd) __pcm_read_lei2d ;
				break ;
		default : return SFE_UNIMPLEMENTED ;
		} ;

	return 0 ;
} /* __wav_open_read */

/*------------------------------------------------------------------------------
 */

int 	__wav_open_write	(SF_PRIVATE *psf)
{	WAV_FMT			wav_fmt ;
	unsigned int	dword ;
	
	if ((psf->sf.format & SF_FORMAT_TYPEMASK) != SF_FORMAT_WAV)
		return	SFE_BAD_OPEN_FORMAT ;
	
	psf->endian      = SF_ENDIAN_LITTLE ;		/* All WAV files are little endian. */
	psf->sf.seekable = SF_TRUE ;
	psf->bytewidth   = BITWIDTH2BYTES (psf->sf.pcmbitwidth) ;
	psf->blockwidth  = psf->bytewidth * psf->sf.channels ;
	psf->error       = 0 ;

	wav_fmt.min.channels    = psf->sf.channels ;
	wav_fmt.min.samplerate  = psf->sf.samplerate ;
	wav_fmt.min.bytespersec = psf->sf.samplerate * psf->bytewidth * psf->sf.channels ;
	wav_fmt.min.blockalign  = psf->bytewidth * psf->sf.channels ;
	wav_fmt.min.bitwidth    = psf->sf.pcmbitwidth ;


	switch (psf->sf.format & SF_FORMAT_SUBMASK)
	{	case	SF_FORMAT_PCM : 
					wav_fmt.format = WAVE_FORMAT_PCM ;

					psf->dataoffset  = 7 * sizeof (dword) + sizeof (MIN_WAV_FMT) ;
					psf->datalength  = psf->blockwidth * psf->sf.samples ;
					psf->filelength  = psf->datalength + psf->dataoffset ;

					write_header_fmt (psf, &wav_fmt, sizeof (MIN_WAV_FMT)) ;
					break ;

		case	SF_FORMAT_FLOAT : 
					wav_fmt.format = WAVE_FORMAT_IEEE_FLOAT ;

					psf->dataoffset  = 7 * sizeof (dword) + sizeof (MIN_WAV_FMT) ;
					psf->datalength  = psf->blockwidth * psf->sf.samples ;
					psf->filelength  = psf->datalength + psf->dataoffset ;

					write_header_fmt (psf, &wav_fmt, sizeof (MIN_WAV_FMT)) ;
					break ;

		case	SF_FORMAT_IMA_ADPCM : 
					wav_fmt.format = WAVE_FORMAT_IMA_ADPCM ;
					ima_writer_init (psf, &wav_fmt) ;

					psf->dataoffset  = 7 * sizeof (dword) + sizeof (IMA_WAV_FMT) ;
					if (psf->sf.samples % wav_fmt.ima.samplesperblock)
						psf->datalength  = ((psf->sf.samples / wav_fmt.ima.samplesperblock) + 1) * wav_fmt.ima.samplesperblock ;
					else
						psf->datalength  = psf->sf.samples ;
					psf->filelength  = psf->datalength + psf->dataoffset ;

					write_header_fmt (psf, &wav_fmt, sizeof (IMA_WAV_FMT)) ;
					break ;

		default : 	return SFE_UNIMPLEMENTED ;
		} ;


	dword = data_MARKER ;
	write (psf->fd, &dword, sizeof (dword)) ;

	dword = H2LE_INT (psf->datalength) ;
	write (psf->fd, &dword, sizeof (dword)) ;
	
	psf->close = (func_close) __wav_close ;

	if ((psf->sf.format & SF_FORMAT_SUBMASK) == SF_FORMAT_FLOAT)
	{	psf->sf.format = (SF_FORMAT_WAV | SF_FORMAT_FLOAT) ;
		if (CAN_READ_WRITE_x86_IEEE)
		{	psf->write_short  = (func_ws) __pcm_write_s2f ;
			psf->write_int    = (func_wi) __pcm_write_i2f ;
			psf->write_double = (func_wd) __pcm_write_d2f ;
			}
		else 
		{	psf->write_short  = (func_ws) wav_write_s2x86f ;
			psf->write_int    = (func_wi) wav_write_i2x86f ;
			psf->write_double = (func_wd) wav_write_d2x86f ;
			} ;
		return	 0 ;
		} ;
	
	if ((psf->sf.format & SF_FORMAT_SUBMASK) == SF_FORMAT_IMA_ADPCM)
	{	psf->sf.format    = (SF_FORMAT_WAV | SF_FORMAT_IMA_ADPCM) ;
		psf->write_short  = (func_ws) ima_write_s ;
		psf->write_int    = (func_wi) ima_write_i ;
		psf->write_double = (func_wd) ima_write_d ;
		psf->seek_func    = (func_seek) ima_seek ;
		psf->close        = (func_close) ima_close ;
		return 0 ;
		} ;

	if ((psf->sf.format & SF_FORMAT_SUBMASK) != SF_FORMAT_PCM)
		return SFE_UNIMPLEMENTED ;

	switch (psf->bytewidth)
	{	case  1 :
				psf->write_short  = (func_ws) __pcm_write_s2uc ;
				psf->write_int    = (func_wi) __pcm_write_i2uc ;
				psf->write_double = (func_wd) __pcm_write_d2uc ;
				break ;
		case  2 :
				psf->write_short  = (func_ws) __pcm_write_s2les ;
				psf->write_int    = (func_wi) __pcm_write_i2les ;
				psf->write_double = (func_wd) __pcm_write_d2les ;
				break ;
		case  3 :
				psf->write_short  = (func_ws) __pcm_write_s2let ;
				psf->write_int    = (func_wi) __pcm_write_i2let ;
				psf->write_double = (func_wd) __pcm_write_d2let ;
				break ;
		case  4 :
				psf->write_short  = (func_ws) __pcm_write_s2lei ;
				psf->write_int    = (func_wi) __pcm_write_i2lei ;
				psf->write_double = (func_wd) __pcm_write_d2lei ;
				break ;
		default : return SFE_UNIMPLEMENTED ;
		} ;

	return 0 ;
} /* wav_open_write */

/*------------------------------------------------------------------------------
 */

static
int	__wav_close	(SF_PRIVATE  *psf)
{	unsigned int		dword ;

	if (psf->mode == SF_MODE_WRITE)
	{	/*  Now we know for certain the length of the file we can
		 *  re-write correct values for the RIFF and data chunks.
		 */

		lseek (psf->fd, 0, SEEK_SET) ;
		psf->filelength = lseek (psf->fd, 0, SEEK_END) ;

		/* Fix RIFF size. */
		dword = H2LE_INT (psf->filelength - 2 * sizeof (dword)) ;
		lseek (psf->fd, sizeof (dword), SEEK_SET) ;
		write (psf->fd, &dword, sizeof (dword)) ;
		
		psf->datalength = psf->filelength - psf->dataoffset ;
		lseek (psf->fd, psf->dataoffset - sizeof (dword), SEEK_SET) ;
		dword = H2LE_INT (psf->datalength) ;
		write (psf->fd, &dword, sizeof (dword)) ;
		} ;

	if (psf->fdata)
		free (psf->fdata) ;
	psf->fdata = NULL ;

	return 0 ;
} /* __wav_close */


/*=========================================================================
 * Private functions.
 */

static
int read_fmt_chunk (SF_PRIVATE *psf, WAV_FMT *wav_fmt)
{	unsigned int	dword, bytesread, structsize  ;
	
	memset (wav_fmt, 0, sizeof (WAV_FMT)) ;
	bytesread = 0 ;

	read (psf->fd, &dword, sizeof (dword)) ;
	structsize = LE2H_INT (dword) ;
	
	__psf_sprintf (psf, "fmt  : %d\n", structsize) ;
	
	if (structsize < 16)
		return SFE_WAV_FMT_SHORT ;
	if (structsize > sizeof (WAV_FMT))
		return SFE_WAV_FMT_TOO_BIG ;
					
	read (psf->fd, wav_fmt, structsize) ;
	bytesread += structsize ;
	if (__CPU_IS_BIG_ENDIAN__)
		le2h_wav_fmt (wav_fmt) ;

	__psf_sprintf (psf, "  Format        : 0x%X => %s\n", wav_fmt->format, wav_format_str (wav_fmt->format)) ;
	__psf_sprintf (psf, "  Channels      : %d\n", wav_fmt->min.channels) ;
	__psf_sprintf (psf, "  Sample Rate   : %d\n", wav_fmt->min.samplerate) ;
	__psf_sprintf (psf, "  Bytes/sec     : %d\n", wav_fmt->min.bytespersec) ;
	__psf_sprintf (psf, "  Block Align   : %d\n", wav_fmt->min.blockalign) ;
	__psf_sprintf (psf, "  Bit Width     : %d\n", wav_fmt->min.bitwidth) ;
	
	psf->sf.samplerate		= wav_fmt->min.samplerate ;
	psf->sf.samples 		= 0 ;					/* Correct this later. */
	psf->sf.channels		= wav_fmt->min.channels ;
	
	switch (wav_fmt->format)
	{	case WAVE_FORMAT_PCM :
		case WAVE_FORMAT_IEEE_FLOAT :
				psf->sf.pcmbitwidth	= wav_fmt->min.bitwidth ;
				psf->bytewidth      = BITWIDTH2BYTES (wav_fmt->min.bitwidth) ;
				break ;

		case WAVE_FORMAT_IMA_ADPCM :
				if (wav_fmt->ima.bitwidth != 4)
					return SFE_WAV_ADPCM_NOT4BIT ;
				if (wav_fmt->ima.channels < 1 || wav_fmt->ima.channels > 2)
					return SFE_WAV_ADPCM_CHANNELS ;
				psf->sf.pcmbitwidth = 16 ;
				psf->bytewidth      = 2 ;
				__psf_sprintf (psf, "  Extra Bytes   : %d\n", wav_fmt->ima.extrabytes) ;
				__psf_sprintf (psf, "  Samples/Block : %d\n", wav_fmt->ima.samplesperblock) ;
				break ;
				
		default : break ;
		} ;
	
	psf->blockwidth = wav_fmt->min.channels * psf->bytewidth ;

	return 0 ;
} /* read_fmt_chunk */

static
int write_header_fmt (SF_PRIVATE *psf, WAV_FMT *wav_fmt, int size)
{	unsigned int dword, RIFFsize ;

	RIFFsize   = psf->filelength - 2 * sizeof (dword) ;

	dword = RIFF_MARKER ;
	write (psf->fd, &dword, sizeof (dword)) ;
	dword = H2LE_INT (RIFFsize) ;
	write (psf->fd, &dword, sizeof (dword)) ;
	
	dword = WAVE_MARKER ;
	write (psf->fd, &dword, sizeof (dword)) ;
	dword = fmt_MARKER ;
	write (psf->fd, &dword, sizeof (dword)) ;
	
	dword = H2LE_INT (size) ;
	write (psf->fd, &dword, sizeof (dword)) ;
	if (__CPU_IS_BIG_ENDIAN__)
		h2le_wav_fmt (wav_fmt) ;	
	write (psf->fd, wav_fmt, size) ;

	return 0 ;
} /* write_header_fmt */

static
const char* wav_format_str (int k)
{	switch (k)
	{	case WAVE_FORMAT_UNKNOWN :
			return "WAVE_FORMAT_UNKNOWN" ;
		case WAVE_FORMAT_PCM          :
			return "WAVE_FORMAT_PCM         " ;
		case WAVE_FORMAT_ADPCM :
			return "WAVE_FORMAT_ADPCM" ;
		case WAVE_FORMAT_IEEE_FLOAT :
			return "WAVE_FORMAT_IEEE_FLOAT" ;
		case WAVE_FORMAT_IBM_CVSD :
			return "WAVE_FORMAT_IBM_CVSD" ;
		case WAVE_FORMAT_ALAW :
			return "WAVE_FORMAT_ALAW" ;
		case WAVE_FORMAT_MULAW :
			return "WAVE_FORMAT_MULAW" ;
		case WAVE_FORMAT_OKI_ADPCM :
			return "WAVE_FORMAT_OKI_ADPCM" ;
		case WAVE_FORMAT_IMA_ADPCM :
			return "WAVE_FORMAT_IMA_ADPCM" ;
		case WAVE_FORMAT_MEDIASPACE_ADPCM :
			return "WAVE_FORMAT_MEDIASPACE_ADPCM" ;
		case WAVE_FORMAT_SIERRA_ADPCM :
			return "WAVE_FORMAT_SIERRA_ADPCM" ;
		case WAVE_FORMAT_G723_ADPCM :
			return "WAVE_FORMAT_G723_ADPCM" ;
		case WAVE_FORMAT_DIGISTD :
			return "WAVE_FORMAT_DIGISTD" ;
		case WAVE_FORMAT_DIGIFIX :
			return "WAVE_FORMAT_DIGIFIX" ;
		case WAVE_FORMAT_DIALOGIC_OKI_ADPCM :
			return "WAVE_FORMAT_DIALOGIC_OKI_ADPCM" ;
		case WAVE_FORMAT_MEDIAVISION_ADPCM :
			return "WAVE_FORMAT_MEDIAVISION_ADPCM" ;
		case WAVE_FORMAT_YAMAHA_ADPCM :
			return "WAVE_FORMAT_YAMAHA_ADPCM" ;
		case WAVE_FORMAT_SONARC :
			return "WAVE_FORMAT_SONARC" ;
		case WAVE_FORMAT_DSPGROUP_TRUESPEECH  :
			return "WAVE_FORMAT_DSPGROUP_TRUESPEECH " ;
		case WAVE_FORMAT_ECHOSC1 :
			return "WAVE_FORMAT_ECHOSC1" ;
		case WAVE_FORMAT_AUDIOFILE_AF18   :
			return "WAVE_FORMAT_AUDIOFILE_AF18  " ;
		case WAVE_FORMAT_APTX :
			return "WAVE_FORMAT_APTX" ;
		case WAVE_FORMAT_AUDIOFILE_AF10   :
			return "WAVE_FORMAT_AUDIOFILE_AF10  " ;
		case WAVE_FORMAT_DOLBY_AC2 :
			return "WAVE_FORMAT_DOLBY_AC2" ;
		case WAVE_FORMAT_GSM610 :
			return "WAVE_FORMAT_GSM610" ;
		case WAVE_FORMAT_MSNAUDIO :
			return "WAVE_FORMAT_MSNAUDIO" ;
		case WAVE_FORMAT_ANTEX_ADPCME :
			return "WAVE_FORMAT_ANTEX_ADPCME" ;
		case WAVE_FORMAT_CONTROL_RES_VQLPC :
			return "WAVE_FORMAT_CONTROL_RES_VQLPC" ;
		case WAVE_FORMAT_DIGIREAL :
			return "WAVE_FORMAT_DIGIREAL" ;
		case WAVE_FORMAT_DIGIADPCM :
			return "WAVE_FORMAT_DIGIADPCM" ;
		case WAVE_FORMAT_CONTROL_RES_CR10 :
			return "WAVE_FORMAT_CONTROL_RES_CR10" ;
		case WAVE_FORMAT_NMS_VBXADPCM :
			return "WAVE_FORMAT_NMS_VBXADPCM" ;
		case WAVE_FORMAT_ROCKWELL_ADPCM :
			return "WAVE_FORMAT_ROCKWELL_ADPCM" ;
		case WAVE_FORMAT_ROCKWELL_DIGITALK :
			return "WAVE_FORMAT_ROCKWELL_DIGITALK" ;
		case WAVE_FORMAT_G721_ADPCM :
			return "WAVE_FORMAT_G721_ADPCM" ;
		case WAVE_FORMAT_MPEG :
			return "WAVE_FORMAT_MPEG" ;
		case WAVE_FORMAT_MPEGLAYER3 :
			return "WAVE_FORMAT_MPEGLAYER3" ;
		case IBM_FORMAT_MULAW :
			return "IBM_FORMAT_MULAW" ;
		case IBM_FORMAT_ALAW :
			return "IBM_FORMAT_ALAW" ;
		case IBM_FORMAT_ADPCM :
			return "IBM_FORMAT_ADPCM" ;
		case WAVE_FORMAT_CREATIVE_ADPCM :
			return "WAVE_FORMAT_CREATIVE_ADPCM" ;
		case WAVE_FORMAT_FM_TOWNS_SND :
			return "WAVE_FORMAT_FM_TOWNS_SND" ;
		case WAVE_FORMAT_OLIGSM :
			return "WAVE_FORMAT_OLIGSM" ;
		case WAVE_FORMAT_OLIADPCM :
			return "WAVE_FORMAT_OLIADPCM" ;
		case WAVE_FORMAT_OLICELP :
			return "WAVE_FORMAT_OLICELP" ;
		case WAVE_FORMAT_OLISBC :
			return "WAVE_FORMAT_OLISBC" ;
		case WAVE_FORMAT_OLIOPR :
			return "WAVE_FORMAT_OLIOPR" ;
		case WAVE_FORMAT_EXTENSIBLE :
			return "WAVE_FORMAT_EXTENSIBLE" ;
		break ;
		} ;
	return "Unknown format" ;
} /* wav_format_str */

/*============================================================================================
** Microsoft 32 bit float (Intel format) Functions.
*/

static
float	read_x86float (unsigned char *cptr)
{	int		exponent, mantissa ;
	float	fvalue ;

	exponent = ((cptr [3] & 0x7F) << 1) | ((cptr [2] & 0x80) ? 1 : 0);
	mantissa = ((cptr [2] & 0x7F) << 16) | (cptr [1] << 8) | (cptr [0]) ;

	if (! (exponent || mantissa))
		return 0.0 ;

	mantissa |= 0x800000 ;
	exponent = exponent ? exponent - 127 : 0 ;
                
	fvalue = mantissa ? ((float) mantissa) / ((float) 0x800000) : 0.0 ;
                
	if (cptr [3] & 0x80)
		fvalue *= -1 ;
                
	if (exponent > 0)
		fvalue *= (1 << exponent) ;
	else if (exponent < 0)
		fvalue /= (1 << abs (exponent)) ;

	return fvalue ;
} /* read_x86float */

static
void	write_x86float (float in, unsigned char *out)
{	int		exponent, mantissa ;

	*((int*) out) = 0 ;
	
	if (in == 0.0)
		return ;
	
	if (in < 0.0)
	{	in *= -1.0 ;
		out [3] |= 0x80 ;
		} ;
		
	in = frexp (in, &exponent) ;
	
	exponent += 126 ;
	
	if (exponent & 0x01)
		out [2] |= 0x80 ;
	out [3] |= (exponent >> 1) & 0x7F ;
	
	in *= (float) 0x1000000 ;
	mantissa = (((int) in) & 0x7FFFFF) ;
	
	out [0]  = mantissa & 0xFF ;
	out [1]  = (mantissa >> 8) & 0xFF ;
	out [2] |= (mantissa >> 16) & 0x7F ;
	
	return ;
} /* write_x86float */

static
void	x86f2s_array (float *buffer, int count, short *ptr, int index)
{	int	k ;
	
	for (k = 0 ; k < count ; k++)
	{	ptr [index] = ((short) read_x86float ((unsigned char *) (buffer +k))) ;
		index ++ ;
		} ;
} /* x86f2s_array */

static
void	x86f2i_array (float *buffer, int count, int *ptr, int index)
{	int	k ;
	
	for (k = 0 ; k < count ; k++)
	{	ptr [index] = ((int) read_x86float ((unsigned char *) (buffer +k))) ;
		index ++ ;
		} ;
} /* x86f2i_array */

static
void	x86f2d_array (float *buffer, double count, double *ptr, int index, double normfact)
{	int	k ;
	
	for (k = 0 ; k < count ; k++)
	{	ptr [index] = ((double) read_x86float ((unsigned char *) (buffer +k))) * normfact  ;
		index ++ ;
		} ;
} /* x86f2d_array */

static 
void	s2x86f_array (short *ptr, int index, float *buffer, int count)
{	int		k ;
	float	value ;
	for (k = 0 ; k < count ; k++)
	{	value = (float) (ptr [index]) ;
		write_x86float (value, (unsigned char*) (buffer + k)) ;
		index ++ ;
		} ;
} /* s2x86f_array */

static 
void	i2x86f_array (int *ptr, int index, float *buffer, int count)
{	int		k ;
	float	value ;
	for (k = 0 ; k < count ; k++)
	{	value = (float) (ptr [index]) ;
		write_x86float (value, (unsigned char*) (buffer + k)) ;
		index ++ ;
		} ;
} /* i2x86f_array */

static 
void	d2x86f_array (double *ptr, int index, float *buffer, int count, double normfact)
{	int		k ;
	float	value ;
	for (k = 0 ; k < count ; k++)
	{	value = (float) (ptr [index] * normfact) ;
		write_x86float (value, (unsigned char*) (buffer + k)) ;
		index ++ ;
		} ;
} /* d2x86f_array */

int		wav_read_x86f2s (SF_PRIVATE *psf, short *ptr, int len)
{	int		bytecount, readcount, bufferlen ;
	int		index = 0, total = 0 ;

	bufferlen = (SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth ;
	bytecount = len * psf->bytewidth ;
	while (bytecount > 0)
	{	readcount = (bytecount >= bufferlen) ? bufferlen : bytecount ;
		total += read (psf->fd, psf->buffer, readcount) ;
		x86f2s_array ((float*) (psf->buffer), readcount / psf->bytewidth, ptr, index) ;
		index += readcount / psf->bytewidth ;
		bytecount -= readcount ;
		} ;

	total /= psf->bytewidth ;
	if (total < len)
		psf->error = SFE_SHORT_READ ;
	
	return total ;
} /* wav_read_x86f2s */

int		wav_read_x86f2i (SF_PRIVATE *psf, int *ptr, int len)
{	int		bytecount, readcount, bufferlen ;
	int		index = 0, total = 0 ;

	bufferlen = (SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth ;
	bytecount = len * psf->bytewidth ;
	while (bytecount > 0)
	{	readcount = (bytecount >= bufferlen) ? bufferlen : bytecount ;
		total += read (psf->fd, psf->buffer, readcount) ;
		x86f2i_array ((float*) (psf->buffer), readcount / psf->bytewidth, ptr, index) ;
		index += readcount / psf->bytewidth ;
		bytecount -= readcount ;
		} ;

	total /= psf->bytewidth ;
	if (total < len)
		psf->error = SFE_SHORT_READ ;
	
	return total ;
} /* wav_read_x86f2i */


int		wav_read_x86f2d (SF_PRIVATE *psf, double *ptr, int len, int normalize)
{	int		bytecount, readcount, bufferlen ;
	int		index = 0, total = 0 ;
	double	normfact ;

	normfact = normalize ? 1.0 : 1.0 ;

	bufferlen = (SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth ;
	bytecount = len * psf->bytewidth ;
	while (bytecount > 0)
	{	readcount = (bytecount >= bufferlen) ? bufferlen : bytecount ;
		total += read (psf->fd, psf->buffer, readcount) ;
		x86f2d_array ((float*) (psf->buffer), readcount / psf->bytewidth, ptr, index, normfact) ;
		index += readcount / psf->bytewidth ;
		bytecount -= readcount ;
		} ;

	total /= psf->bytewidth ;
	if (total < len)
		psf->error = SFE_SHORT_READ ;
	
	return total ;
} /* wav_read_x86f2d */

static 
int	wav_write_s2x86f (SF_PRIVATE *psf, short *ptr, int len)
{	int		bytecount, writecount, bufferlen ;
	int		index = 0, total = 0 ;

	bufferlen = (SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth ;
	bytecount = len * psf->bytewidth ;
	while (bytecount > 0)
	{	writecount = (bytecount >= bufferlen) ? bufferlen : bytecount ;
		s2x86f_array (ptr, index, (float*) (psf->buffer), writecount / psf->bytewidth) ;
		total += write (psf->fd, psf->buffer, writecount) ;
		index += writecount / psf->bytewidth ;
		bytecount -= writecount ;
		} ;

	total /= psf->bytewidth ;
	if (total < len)
		psf->error = SFE_SHORT_WRITE ;
	
	return total ;
} /* wav_write_s2x86f */

static 
int	wav_write_i2x86f (SF_PRIVATE *psf, int *ptr, int len) 
{	int		bytecount, writecount, bufferlen ;
	int		index = 0, total = 0 ;

	bufferlen = (SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth ;
	bytecount = len * psf->bytewidth ;
	while (bytecount > 0)
	{	writecount = (bytecount >= bufferlen) ? bufferlen : bytecount ;
		i2x86f_array (ptr, index, (float*) (psf->buffer), writecount / psf->bytewidth) ;
		total += write (psf->fd, psf->buffer, writecount) ;
		index += writecount / psf->bytewidth ;
		bytecount -= writecount ;
		} ;

	total /= psf->bytewidth ;
	if (total < len)
		psf->error = SFE_SHORT_WRITE ;
	
	return total ;
} /* wav_write_i2x86f */

static 
int	wav_write_d2x86f (SF_PRIVATE *psf, double *ptr, int len, int normalize)
{	int		bytecount, writecount, bufferlen ;
	int		index = 0, total = 0 ;
	double		normfact ;
	
	normfact = (normalize) ? 1.0 : 1.0 ;

	bufferlen = (SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth ;
	bytecount = len * psf->bytewidth ;
	while (bytecount > 0)
	{	writecount = (bytecount >= bufferlen) ? bufferlen : bytecount ;
		d2x86f_array (ptr, index, (float*) (psf->buffer), writecount / psf->bytewidth, normfact) ;
		total += write (psf->fd, psf->buffer, writecount) ;
		index += writecount / psf->bytewidth ;
		bytecount -= writecount ;
		} ;

	total /= psf->bytewidth ;
	if (total < len)
		psf->error = SFE_SHORT_WRITE ;
	
	return total ;
} /* wav_write_d2x86f */


/*============================================================================================
** IMA ADPCM Read Functions.
*/

static int ima_index_adjust [16] = 
{	-1, -1, -1, -1,		/* +0 - +3, decrease the step size */
     2,  4,  6,  8,     /* +4 - +7, increase the step size */
    -1, -1, -1, -1,		/* -0 - -3, decrease the step size */
     2,  4,  6,  8,		/* -4 - -7, increase the step size */
} ;

static int ima_step_size [89] = 
{	7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 19, 21, 23, 25, 28, 31, 34, 37, 41, 45, 
	50, 55, 60, 66, 73, 80, 88, 97, 107, 118, 130, 143, 157, 173, 190, 209, 230, 
	253, 279, 307, 337, 371, 408, 449, 494, 544, 598, 658, 724, 796, 876, 963, 
	1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066, 2272, 2499, 2749, 3024, 3327, 
	3660, 4026, 4428, 4871, 5358, 5894, 6484, 7132, 7845, 8630, 9493, 10442,
	11487, 12635, 13899, 15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 
	32767
} ;

static
int	ima_reader_init (SF_PRIVATE *psf, WAV_FMT *fmt)
{	IMA_PRIVATE	*pima ;
	int 		count, pimasize ;
	
	pimasize = sizeof (IMA_PRIVATE) + fmt->ima.blockalign + 3 * fmt->ima.channels * fmt->ima.samplesperblock ;

	if (! (psf->fdata = malloc (pimasize)))
		return SFE_MALLOC_FAILED ;
	pima = (IMA_PRIVATE*) psf->fdata ;
	memset (pima, 0, pimasize) ;
	
	pima->block   = (unsigned char*) pima->data ;
	pima->samples = (short*) (pima->data + fmt->ima.blockalign) ;
	
	pima->channels        = fmt->ima.channels ;
	pima->blocksize       = fmt->ima.blockalign ;
	pima->samplesperblock = fmt->ima.samplesperblock ;

	if (psf->datalength % pima->blocksize)
	{	__psf_sprintf (psf, "*** Warning : data chunk seems to be truncated.\n") ;
		pima->blocks = psf->datalength / pima->blocksize  + 1 ;
		}
	else
		pima->blocks = psf->datalength / pima->blocksize ;
		
	count = 2 * (pima->blocksize - 4 * pima->channels) / pima->channels + 1 ;
	if (pima->samplesperblock != count)
		__psf_sprintf (psf, "*** Warning : samplesperblock shoud be %d.\n", count) ;

	psf->sf.samples = pima->samplesperblock * pima->blocks ;
	pima->samplecount = pima->samplesperblock ;
	
	return 0 ;
} /* ima_reader_init */

static
int		ima_read_block (SF_PRIVATE *psf, IMA_PRIVATE *pima)
{	int		chan, k, current, blockindex, index, indexstart ;
	short	step, diff, bytecode, stepindex [2] ;
	
	if ((k = read (psf->fd, pima->block, pima->blocksize)) != pima->blocksize)
		__psf_sprintf (psf, "*** Warning : short read (%d != %d).\n", k, pima->blocksize) ;

	pima->samplecount = 0 ;
		
	/* Read and check the block header. */
	
	for (chan = 0 ; chan < pima->channels ; chan++)
	{	current = pima->block [chan*4] | (pima->block [chan*4+1] << 8) ;
		if (current & 0x8000)
			current -= 0x10000 ;
			
		stepindex [chan] = pima->block [chan*4+2] ;
		if (stepindex [chan] < 0)
			stepindex [chan] = 0 ;
		else if (stepindex [chan] > 88)
			stepindex [chan] = 88 ;

		if (pima->block [chan*4+3] != 0)
			__psf_sprintf (psf, "IMA ADPCM synchanonisation error.\n") ;
		
		pima->samples [chan] = current ;

		/* __psf_sprintf (psf, "block %d : channel %d (current, index) : (%d, %d)\n", 
		**	pima->blockcount, chan,  current, stepindex [chan]) ;
		*/
		} ;
		
	/* Pull apart the packed 4 bit samples and store them in their
	** correct sample positions.
	*/
	
	blockindex = 4 * pima->channels ;
	
	indexstart = pima->channels ;
	while (blockindex <  pima->blocksize)
	{	for (chan = 0 ; chan < pima->channels ; chan++)
		{	index = indexstart + chan ;
			for (k = 0 ; k < 4 ; k++)
			{	bytecode = pima->block [blockindex++] ;
				pima->samples [index] = bytecode & 0x0F ;
				index += pima->channels ;
				pima->samples [index] = (bytecode >> 4) & 0x0F ;
				index += pima->channels ;
				} ;
			} ;
		indexstart += 8 * pima->channels ;
		} ;
		
	/* Decode the encoded 4 bit samples. */
	
	for (k = pima->channels ; k < (pima->samplesperblock * pima->channels) ; k ++)
	{	chan = (pima->channels > 1) ? (k % 2) : 0 ;

		bytecode = pima->samples [k] & 0xF ;
		
		step = ima_step_size [stepindex [chan]] ;
		current = pima->samples [k - pima->channels] ;
  
		diff = step >> 3 ;
		if (bytecode & 1) 
			diff += step >> 2 ;
		if (bytecode & 2) 
			diff += step >> 1 ;
		if (bytecode & 4) 
			diff += step ;
		if (bytecode & 8) 
			diff = -diff ;

		current += diff ;

		if (current > 32767) 
			current = 32767;
		else if (current < -32768) 
			current = -32768 ;

		stepindex [chan] += ima_index_adjust [bytecode] ;
	
		if (stepindex [chan] < 0) 
			stepindex [chan] = 0 ;
		else if (stepindex [chan] > 88) 
			stepindex [chan] = 88 ;

		pima->samples [k] = current ;
		} ;

	pima->blockcount ++ ;
	
	return 1 ;
} /* ima_read_block */

static
int ima_read (SF_PRIVATE *psf, IMA_PRIVATE *pima, short *ptr, int len)
{	int		count, total = 0, index = 0 ;
	
	while (index < len)
	{	if (pima->blockcount >= pima->blocks && pima->samplecount >= pima->samplesperblock)
		{	memset (&(ptr[index]), 0, (len - index) * sizeof (short)) ;
			return total ;
			} ;
		
		if (pima->samplecount >= pima->samplesperblock)
			ima_read_block (psf, pima) ;
		
		count = (pima->samplesperblock - pima->samplecount) * pima->channels ;
		count = (len - index > count) ? count : len - index ;
		
		memcpy (&(ptr[index]), &(pima->samples [pima->samplecount]), count * sizeof (short)) ;
		index += count ;
		pima->samplecount += count / pima->channels ;
		total = index ;
		} ;

	return total ;		
} /* ima_read */

static
int		ima_read_s (SF_PRIVATE *psf, short *ptr, int len)
{	IMA_PRIVATE 	*pima ; 
	int				total ;

	if (! psf->fdata)
		return 0 ;
	pima = (IMA_PRIVATE*) psf->fdata ;
	
	total = ima_read (psf, pima, ptr, len) ;

	return total ;
} /* ima_read_s */

static
int		ima_read_i  (SF_PRIVATE *psf, int *ptr, int len)
{	IMA_PRIVATE *pima ; 
	short		*sptr ;
	int			k, bufferlen, readcount = 0, count ;
	int			index = 0, total = 0 ;

	if (! psf->fdata)
		return 0 ;
	pima = (IMA_PRIVATE*) psf->fdata ;
	
	sptr = (short*) psf->buffer ;
	bufferlen = ((SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth) / sizeof (short) ;
	while (len > 0)
	{	readcount = (len >= bufferlen) ? bufferlen : len ;
		count = ima_read (psf, pima, sptr, readcount) ;
		for (k = 0 ; k < readcount ; k++)
			ptr [index+k] = (int) (sptr [k]) ;
		index += readcount ;
		total += count ;
		len -= readcount ;
		} ;
	return total ;
} /* ima_read_i */

static
int		ima_read_d  (SF_PRIVATE *psf, double *ptr, int len, int normalize)
{	IMA_PRIVATE *pima ; 
	short		*sptr ;
	int			k, bufferlen, readcount = 0, count ;
	int			index = 0, total = 0 ;

	if (! psf->fdata)
		return 0 ;
	pima = (IMA_PRIVATE*) psf->fdata ;
	
	sptr = (short*) psf->buffer ;
	bufferlen = ((SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth) / sizeof (short) ;
	while (len > 0)
	{	readcount = (len >= bufferlen) ? bufferlen : len ;
		count = ima_read (psf, pima, sptr, readcount) ;
		for (k = 0 ; k < readcount ; k++)
			ptr [index+k] = (double) (sptr [k]) ;
		index += readcount ;
		total += count ;
		len -= readcount ;
		} ;
	return total ;
} /* ima_read_d */


static 
off_t    ima_seek   (SF_PRIVATE *psf, off_t offset, int whence)
{	IMA_PRIVATE *pima ; 
	int			newblock, newsample ;
	
	if (! psf->fdata)
		return 0 ;
	pima = (IMA_PRIVATE*) psf->fdata ;

	if (! (psf->blockwidth && psf->datalength && psf->dataoffset))
	{	psf->error = SFE_BAD_SEEK ;
		return	((off_t) -1) ;
		} ;
		
	switch (whence)
	{	case SEEK_SET :
				if (offset < 0 || offset > pima->blocks * pima->samplesperblock)
				{	psf->error = SFE_BAD_SEEK ;
					return	((off_t) -1) ;
					} ;
				newblock  = offset / pima->samplesperblock ;
				newsample = offset % pima->samplesperblock ;
				break ;
				
		case SEEK_CUR :
				if (psf->current + offset < 0 || psf->current + offset > pima->blocks * pima->samplesperblock)
				{	psf->error = SFE_BAD_SEEK ;
					return	((off_t) -1) ;
					} ;
				newblock  = (psf->current + offset) / pima->samplesperblock ;
				newsample = (psf->current + offset) % pima->samplesperblock ;
				break ;
				
		case SEEK_END :
				if (offset > 0 || pima->samplesperblock * pima->blocks + offset < 0)
				{	psf->error = SFE_BAD_SEEK ;
					return	((off_t) -1) ;
					} ;
				newblock  = (pima->samplesperblock * pima->blocks + offset) / pima->samplesperblock ;
				newsample = (pima->samplesperblock * pima->blocks + offset) % pima->samplesperblock ;
				break ;
				
		default : 
				psf->error = SFE_BAD_SEEK ;
				return	((off_t) -1) ;
		} ;
		
	if (psf->mode == SF_MODE_READ)
	{	lseek (psf->fd, psf->dataoffset + newblock * pima->blocksize, SEEK_SET) ;
		ima_read_block (psf, pima) ;
	
		pima->blockcount  = newblock ;
		pima->samplecount = newsample ;
		}
	else
	{	/* What to do about write??? */ 
		psf->error = SFE_BAD_SEEK ;
		return	((off_t) -1) ;
		} ;

	psf->current = newblock * pima->samplesperblock + newsample ;
	return psf->current ;
} /* ima_seek */

/*==========================================================================================
** IMA ADPCM Write Functions.
*/

static
int	ima_writer_init (SF_PRIVATE *psf, WAV_FMT *fmt)
{	IMA_PRIVATE	*pima ;
	int 		pimasize ;
	
	if (fmt->format != WAVE_FORMAT_IMA_ADPCM)
		__psf_sprintf (psf, "*** Warning : format tag != WAVE_FORMAT_IMA_ADPCM.\n") ;
	
	fmt->ima.bitwidth        = 4 ;
	fmt->ima.extrabytes      = 2 ;
	fmt->ima.blockalign      = 1024 ;
	fmt->ima.samplesperblock = 2 * (fmt->ima.blockalign - 4 * fmt->ima.channels) / fmt->ima.channels + 1 ;

	pimasize = sizeof (IMA_PRIVATE) + fmt->ima.blockalign + 3 * fmt->ima.channels * fmt->ima.samplesperblock ;

	if (! (psf->fdata = malloc (pimasize)))
		return SFE_MALLOC_FAILED ;
	pima = (IMA_PRIVATE*) psf->fdata ;
	memset (pima, 0, pimasize) ;
	
	pima->channels        = fmt->ima.channels ;
	pima->blocksize       = fmt->ima.blockalign ;
	pima->samplesperblock = fmt->ima.samplesperblock ;

	pima->block   = (unsigned char*) pima->data ;
	pima->samples = (short*) (pima->data + fmt->ima.blockalign) ;
	
	pima->samplecount = 0 ;
	
	return 0 ;
} /* ima_writer_init */


/*==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
**==========================================================================================
*/

static
int		ima_write_block (SF_PRIVATE *psf, IMA_PRIVATE *pima)
{	int		chan, k, step, diff, vpdiff, blockindex, index, indexstart ;
	short	bytecode, mask ;
	
	/* Write the block header. */

	for (chan = 0 ; chan < pima->channels ; chan++)
	{	pima->block [chan*4]   = pima->samples [chan] & 0xFF ;
		pima->block [chan*4+1] = (pima->samples [chan] >> 8) & 0xFF ;
		
		pima->block [chan*4+2] = pima->stepindex [chan] ;
		pima->block [chan*4+3] = 0 ;
		} ;
		
	pima->previous  [0] = pima->samples [0] ;
	pima->previous  [1] = pima->samples [1] ;

	/* Encode the samples as 4 bit. */

	for (k = pima->channels ; k < (pima->samplesperblock * pima->channels) ; k ++)
	{	chan = (pima->channels > 1) ? (k % 2) : 0 ;
	
		diff = pima->samples [k] - pima->previous [chan] ;
	
		bytecode = 0 ;
		step = ima_step_size [pima->stepindex [chan]] ;
		vpdiff = step >> 3 ;
		if (diff < 0) 
		{	bytecode = 8 ; 
			diff = -diff ;
			} ;
		mask = 4 ;
		while (mask)
		{	if (diff >= step)
			{	bytecode |= mask ;
				diff -= step ;
				vpdiff += step ;
				} ;
			step >>= 1 ;
			mask >>= 1 ;
			} ;

		if (bytecode & 8)
			pima->previous [chan] -= vpdiff ;
		else
			pima->previous [chan] += vpdiff ;

		if (pima->previous [chan] > 32767) 
			pima->previous [chan] = 32767;
		else if (pima->previous [chan] < -32768) 
			pima->previous [chan] = -32768;

		pima->stepindex [chan] += ima_index_adjust [bytecode] ;
		if (pima->stepindex [chan] < 0)
			pima->stepindex [chan] = 0 ;
		else if (pima->stepindex [chan] > 88)
			pima->stepindex [chan] = 88 ;
		
		pima->samples [k] = bytecode ;
		} ;

	/* Pack the 4 bit encoded samples. */

	blockindex = 4 * pima->channels ;
	
	indexstart = pima->channels ;
	while (blockindex <  pima->blocksize)
	{	for (chan = 0 ; chan < pima->channels ; chan++)
		{	index = indexstart + chan ;
			for (k = 0 ; k < 4 ; k++)
			{	pima->block [blockindex] = pima->samples [index] & 0x0F ;
				index += pima->channels ;
				pima->block [blockindex] |= (pima->samples [index] << 4) & 0xF0 ;
				index += pima->channels ;
				blockindex ++ ;
				} ;
			} ;
		indexstart += 8 * pima->channels ;
		} ;
		
	/* Write the block to disk. */

	if ((k = write (psf->fd, pima->block, pima->blocksize)) != pima->blocksize)
		__psf_sprintf (psf, "*** Warning : short write (%d != %d).\n", k, pima->blocksize) ;
		
	memset (pima->samples, 0, pima->samplesperblock * sizeof (short)) ;
	pima->samplecount = 0 ;
			
	return 1 ;
} /* ima_write_block */

static
int ima_write (SF_PRIVATE *psf, IMA_PRIVATE *pima, short *ptr, int len)
{	int		count, total = 0, index = 0 ;
	
	while (index < len)
	{	count = (pima->samplesperblock - pima->samplecount) * pima->channels ;
		if (count > len)
			count = len ;
		else if (count > len - index)
			count = len - index ;

		memcpy (&(pima->samples [pima->samplecount]), &(ptr[index]), count * sizeof (short)) ;
		index += count ;
		pima->samplecount += count / pima->channels ;
		total = index ;

		if (pima->samplecount >= pima->samplesperblock)
			ima_write_block (psf, pima) ;	
		} ;

	return total ;		
} /* ima_write */

static
int		ima_write_s (SF_PRIVATE *psf, short *ptr, int len)
{	IMA_PRIVATE 	*pima ; 
	int				total ;

	if (! psf->fdata)
		return 0 ;
	pima = (IMA_PRIVATE*) psf->fdata ;
	
	total = ima_write (psf, pima, ptr, len) ;

	return total ;
} /* ima_write_s */

static
int		ima_write_i  (SF_PRIVATE *psf, int *ptr, int len)
{	IMA_PRIVATE *pima ; 
	short		*sptr ;
	int			k, bufferlen, writecount = 0, count ;
	int			index = 0, total = 0 ;

	if (! psf->fdata)
		return 0 ;
	pima = (IMA_PRIVATE*) psf->fdata ;
	
	sptr = (short*) psf->buffer ;
	bufferlen = ((SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth) / sizeof (short) ;
	while (len > 0)
	{	writecount = (len >= bufferlen) ? bufferlen : len ;
		count = ima_write (psf, pima, sptr, writecount) ;
		for (k = 0 ; k < writecount ; k++)
			ptr [index+k] = (int) (sptr [k]) ;
		index += writecount ;
		total += count ;
		len -= writecount ;
		} ;
	return total ;
} /* ima_write_i */

static
int		ima_write_d  (SF_PRIVATE *psf, double *ptr, int len, int normalize)
{	IMA_PRIVATE *pima ; 
	short		*sptr ;
	int			k, bufferlen, writecount = 0, count ;
	int			index = 0, total = 0 ;

	if (! psf->fdata)
		return 0 ;
	pima = (IMA_PRIVATE*) psf->fdata ;
	
	sptr = (short*) psf->buffer ;
	bufferlen = ((SF_BUFFER_LEN / psf->blockwidth) * psf->blockwidth) / sizeof (short) ;
	while (len > 0)
	{	writecount = (len >= bufferlen) ? bufferlen : len ;
		count = ima_write (psf, pima, sptr, writecount) ;
		for (k = 0 ; k < writecount ; k++)
			ptr [index+k] = (double) (sptr [k]) ;
		index += writecount ;
		total += count ;
		len -= writecount ;
		} ;
	return total ;
} /* ima_write_d */

static
int	ima_close	(SF_PRIVATE  *psf)
{	IMA_PRIVATE *pima ; 
	unsigned int		dword ;

	if (! psf->fdata)
	{	__wav_close (psf) ;
		return 0 ;
		} ;
	pima = (IMA_PRIVATE*) psf->fdata ;

	if (psf->mode == SF_MODE_WRITE)
	{	/*  Now we know for certain the length of the file we can
		**  re-write correct values for the RIFF and data chunks.
		*/
		 
		if (pima->samplecount < pima->samplesperblock)
			ima_write_block (psf, pima) ;	

		lseek (psf->fd, 0, SEEK_SET) ;
		psf->filelength = lseek (psf->fd, 0, SEEK_END) ;

		/* Fix RIFF size. */
		dword = H2LE_INT (psf->filelength - 2 * sizeof (dword)) ;
		lseek (psf->fd, sizeof (dword), SEEK_SET) ;
		write (psf->fd, &dword, sizeof (dword)) ;
		
		psf->datalength = psf->filelength - psf->dataoffset ;
		lseek (psf->fd, psf->dataoffset - sizeof (dword), SEEK_SET) ;
		dword = H2LE_INT (psf->datalength) ;
		write (psf->fd, &dword, sizeof (dword)) ;
		} ;

	if (psf->fdata)
		free (psf->fdata) ;
	psf->fdata = NULL ;

	return 0 ;
} /* ima_close */


