/*
** Copyright (C) 1999 Erik de Castro Lopo <erikd@zip.com.au>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/



#include	<stdio.h>
#include	<string.h>
#include	<unistd.h>
#include	<math.h>

#include	<sndfile.h>

#define	BUFFER_SIZE		(1<<15)

#define		ADPCM_MAP(x) 	(((x) < (BUFFER_SIZE / 4)) ? (x) : ((BUFFER_SIZE/2) -(x)))

static	void	adpcm_test_short	(char *str, char *filename, int typemajor, int typeminor) ;

/* Force the start of this buffer to be double aligned. Sparc-solaris will
** choke if its not.
*/
static	double	test_buffer [(BUFFER_SIZE/sizeof(double))+1] ;

int		main (int argc, char *argv[])
{	char	*filename ;

	if (argc != 2)
	{	printf ("Usage : %s <test>\n", argv [0]) ;
		printf ("    Where <test> is one of the following:\n") ;
		printf ("           wav_ima  - test WAV IAM ADPCM file functions\n") ;
		exit (1) ;
		} ;
		
	if (! strcmp (argv [1], "wav_ima"))
	{	filename = "test.wav" ;
		adpcm_test_short	("wav_ima", filename, SF_FORMAT_WAV, SF_FORMAT_IMA_ADPCM) ;
		unlink (filename) ;
		return 0 ;
		} ;

	printf ("************************************\n") ;
	printf ("*  No '%s' test defined.\n", argv [1]) ;
	printf ("************************************\n") ;
	return 1 ;
} /* main */

/*============================================================================================
 *	Here are the test functions.
 */ 
 
static	
void	adpcm_test_short (char *str, char *filename, int typemajor, int typeminor)
{	SNDFILE		*file ;
	SF_INFO		sfinfo ;
	int			k, m ,datalen ;
	short		*data ;

	printf ("    adpcm_test_short : %s ... ", str) ;
	
	datalen = BUFFER_SIZE / sizeof (short) ;

	data = (short*) test_buffer ;
	for (k = 0 ; k < datalen ; k++)
		data [k] = ADPCM_MAP (k) ;
		
	sfinfo.samplerate  = 44100 ;
	sfinfo.samples     = datalen ;
	sfinfo.channels    = 1 ;
	sfinfo.pcmbitwidth = 16 ;
	sfinfo.format 	   = (typemajor | typeminor) ;

	if (! (file = sf_open_write (filename, &sfinfo)))
	{	printf ("sf_open_write failed with error : ") ;
		sf_perror (NULL) ;
		exit (1) ;
		} ;
	
	if ((k = sf_write_short (file, data, datalen)) != datalen)
	{	printf ("sf_write_short failed with short write (%d => %d).\n", datalen, k) ;
		exit (1) ;
		} ;
	sf_close (file) ;
	
	memset (data, 0, datalen * sizeof (short)) ;
	memset (&sfinfo, 0, sizeof (sfinfo)) ;
	
	if (! (file = sf_open_read (filename, &sfinfo)))
	{	printf ("sf_open_read failed with error : ") ;
		sf_perror (NULL) ;
		exit (1) ;
		} ;
	
	if (sfinfo.format != (typemajor | typeminor))
	{	printf ("Returned format incorrect (0x%08X => 0x%08X).\n", (typemajor | typeminor), sfinfo.format) ;
		exit (1) ;
		} ;
	
	if (sfinfo.samples < datalen)
	{	printf ("Too few samples in file. (%d => %d)\n", datalen, sfinfo.samples) ;
		exit (1) ;
		} ;
	
	if (sfinfo.channels != 1)
	{	printf ("Incorrect number of channels in file.\n") ;
		exit (1) ;
		} ;

	if (sfinfo.pcmbitwidth != 16)
	{	printf ("Incorrect bit width (%d).\n", sfinfo.pcmbitwidth) ;
		exit (1) ;
		} ;

	if ((k = sf_read_short (file, data, datalen)) != datalen)
	{	printf ("short read (%d).\n", k) ;
		exit (1) ;
		} ;

	for (k = 0 ; k < datalen ; k++)
		if (abs (data [k] - ADPCM_MAP (k)) > 1)
		{	printf ("Incorrect sample (A #%d : %d => %d).\n", k, ADPCM_MAP (k), data [k]) ;
			exit (1) ;
			} ;

	if ((k = sf_read_short (file, data, datalen)) != sfinfo.samples - datalen)
	{	printf ("Incorrect read length (%d => %d).\n", sfinfo.samples - datalen, k) ;
		exit (1) ;
		} ;
		
	for (k = 0 ; k < sfinfo.samples - datalen ; k++)
		if (data [k])
		{	printf ("Incorrect sample (B #%d : %d != 0).\n", datalen + k, data [k]) ;
			exit (1) ;
			} ;

	/* Now test sf_seek function. */
	
	if ((k = sf_seek (file, 0, SEEK_SET)) != 0)
	{	printf ("Seek to start of file failed (%d).\n", k) ;
		exit (1) ;
		} ;

	for (m = 0 ; m < 3 ; m++)
	{	if ((k = sf_read_short (file, data, datalen/7)) != datalen / 7)
		{	printf ("Incorrect read length (%d => %d).\n", datalen / 7, k) ;
			exit (1) ;
			} ;

		for (k = 0 ; k < datalen/7 ; k++)
			if (data [k] != ADPCM_MAP (k + m * (datalen / 7)))
			{	printf ("Incorrect sample (C #%d : %d != 0).\n", k + m * (datalen / 7), data [k]) ;
				exit (1) ;
				} ;
		} ;

	/* Now test sf_seek function. */
	
	
	/* Check seek from start of file. */
	if ((k = sf_seek (file, 1, SEEK_SET)) != 1)
	{	printf ("Seek to start of file + 1 failed (%d).\n", k) ;
		exit (1) ;
		} ;
	if ((k = sf_read_short (file, data, 1)) != 1)
	{	printf ("sf_read_short (file, data, 1) returned %d.\n", k) ;
		exit (1) ;
		} ;
	
	if (data [0] != ADPCM_MAP (1))
	{	printf ("sf_seek (SEEK_SET) followed by sf_read_short failed (%d, %d).\n", ADPCM_MAP (1), data [0]) ;
		exit (1) ;
		} ;
	
	k = sf_seek (file, 2, SEEK_CUR) ;
	sf_read_short (file, data, 1) ;
	if (data [0] != ADPCM_MAP (4) || k != 4)
	{	printf ("sf_seek (SEEK_CUR) followed by sf_read_short failed (%d, %d).\n", data [0], k) ;
		exit (1) ;
		} ;
	
	/* Check seek backward from current position. */
	k = sf_seek (file, -2, SEEK_CUR) ;
	sf_read_short (file, data, 1) ;
	if (data [0] != ADPCM_MAP (3) || k != 3)
	{	printf ("sf_seek (SEEK_CUR) followed by sf_read_short failed (%d, %d).\n", data [0], k) ;
		exit (1) ;
		} ;
	
	/* Check that read past end of file returns number of items. */
	sf_seek (file, datalen, SEEK_SET) ;

 	if ((k = sf_read_short (file, data, datalen)) != sfinfo.samples - datalen)
 	{	printf ("Return value from sf_read_short past end of file incorrect (%d).\n", k) ;
 		exit (1) ;
 		} ;
	
	/* Check seek backward from end. */
	k = sf_seek (file, 5 - sfinfo.samples, SEEK_END) ;
	sf_read_short (file, data, 1) ;
	if (data [0] != ADPCM_MAP (5) || k != 5)
	{	printf ("sf_seek (SEEK_END) followed by sf_read_short failed (%d, %d).\n", data [0], k) ;
		exit (1) ;
		} ;

	sf_close (file) ;

	printf ("ok\n") ;
} /* adpcm_test_short */


